//: blocco_jacobi.h Classe rappresentante un blocco adatto all'esecuzione di Jacobi.
//. Consente di rappresentare un blocco di numeri in virgola mobile
//. sotto forma di matrice.
//. Consente il calcolo dell'iterazione di Jacobi anche ai limiti.

#ifndef _BLOCCO_JACOBI
#define _BLOCCO_JACOBI

#include <stdlib.h>
#include <math.h>
#include <vector.h>
#include <iostream.h>
#include <time.h>

float max(float a, float b) {
	if (a>=b)
		return a;
	else
		return b;
}

class BloccoJacobi {
	protected:
	vector< vector< float > > it_att; 	// matrice dell'iterazione attuale.
	vector< vector < float > > it_new; 	// matrice contenente i nuovi valori.
										// Da notare che va indicizzata normalmente
										// ossia parte da 1 e non da 0.
	int ux, uy, dx, dy;	// (ux, uy)  = angolo alto sinistro (dx, dy) = angolo basso
						// destro delle matrici.
	unsigned int righe, colonne;	// numero di righe e colonne delle matrici
	float tempo_calcolo;	// tempo impiegato per il calcolo dell'algoritmo.
	float convergenza;	// errore relativo desiderato.

	public:
	BloccoJacobi(int ux, int uy, int dx, int dy) :
		ux(ux), uy(uy), dx(dx), dy(dy),
		righe(uy - dy + 1), colonne(dx - ux + 1) {
		
		tempo_calcolo = 0.0;
		convergenza = 0.1;

		// Preparo le matrici che contengono il blocco
		it_att.resize(righe+1);
		it_new.resize(righe+1);
		if ( (it_att.size()!=righe+1) || (it_new.size()!=righe+1) ) {
			cerr << "Errore durante la creazione del blocco della matricie"
			     << "( (" << ux << "," << uy << "), (" << dx << "," << dy << ") )";
			exit(1);
		}
		for (unsigned int i=0; i<it_att.size(); i++) {
			it_att[i].resize(colonne+1);
			it_new[i].resize(colonne+1);
			if ( (it_att[i].size()!=colonne+1) || (it_new[i].size()!=colonne+1) ) {
				cerr << "Errore durante la creazione del blocco della matricie"
			     << "( (" << ux << "," << uy << "), (" << dx << "," << dy << ") )";
				exit(1);
			}
		}
	}
	
	BloccoJacobi(unsigned int righe, unsigned int colonne) : 
		righe(righe), colonne(colonne) {
		ux = 1;
		dy = 1;
		dx = colonne;
		uy = righe;

		// Preparo le matrici che contengono il blocco
		it_att.resize(righe+1);
		it_new.resize(righe+1);
		if ( (it_att.size()!=righe+1) || (it_new.size()!=righe+1) ) {
			cerr << "Errore durante la creazione del blocco della matricie"
			     << "( (" << ux << "," << uy << "), (" << dx << "," << dy << ") )";
			exit(1);
		}
		
		for (unsigned int i=0; i<it_att.size(); i++) {
			it_att[i].resize(colonne+1);
			it_new[i].resize(colonne+1);
			if ( (it_att[i].size()!=colonne+1) || (it_new[i].size()!=colonne+1) ) {
				cerr << "Errore durante la creazione del blocco della matricie"
			         << "( (" << ux << "," << uy << "), (" << dx << "," << dy << ") )";
				exit(1);
			}
		}
	}
	
	void set_convergenza(float c) {
		if (c>0) convergenza = c;
	}
	
	void set(int riga, int colonna, float v) {
		if( riga>=uy && riga<=dy ) {
			cerr << "Set. Indice riga fuore dai limiti" << endl;
			return;
		}
		if( colonna>=dx && colonna<=ux ) {
			cerr << "Set. Indice colonna fuori dai limiti" << endl;
			return;
		}
		it_att[riga][colonna] = v;
	}
	
	void set_all(float v) {
		for (unsigned int i=1; i<=righe; i++)
			for (unsigned int k=1; k<=colonne; k++) {
				it_att[i][k] = v;
				it_new[i][k] = v;
			}
	}
	
	void ai_limiti(float nord, float sud, float est, float ovest) {
		for (unsigned int i=1; i<=righe; i++) {
			it_att[i][1] = est;
			it_new[i][1] = est;
		}
	
		for (unsigned int i=1; i<=righe; i++) {
			it_att[i][colonne] = ovest;
			it_new[i][colonne] = ovest;
		}
	
		for (unsigned int i=1; i<=colonne; i++) {
			it_att[1][i] = nord;
			it_new[1][i] = nord;
		}
		
		for (unsigned int i=1; i<=colonne; i++) {
			it_att[righe][i] = sud;
			it_new[righe][i] = sud;
		}
	}
	
	unsigned int calcola_jacobi() {
		float errore_relativo;
		unsigned int n = 0;
		
		tempo_calcolo = clock();
		
		do {
			for (unsigned int i=2; i<righe; i++)
				for (unsigned int j=2 ; j<colonne; j++) {
					it_new[i][j] = 0.25*(it_att[i-1][j] + it_att[i][j+1] + it_att[i+1][j] + it_att[i][j-1]);
				}
				
			errore_relativo = 0.0;
			for (unsigned int i=1; i<=righe; i++)
				for (unsigned int j=1 ; j<=colonne; j++) {
					errore_relativo = max(errore_relativo, fabs(it_new[i][j] - it_att[i][j]));
					it_att[i][j] = it_new[i][j];
				}
				
			n++;
			
		#ifdef DEBUG
		cout << ".";
		#endif
		} while (errore_relativo > convergenza);
		
		#ifdef DEBUG
		cout << endl;
		#endif
		
		tempo_calcolo = (clock() - tempo_calcolo) / CLOCKS_PER_SEC;
		
		return n;
	}
	
	float get_tempo_calcolo() { return tempo_calcolo; }
	
	void mostra_risultato() {
		for (unsigned int i=1; i<=righe; i++) {
			for (unsigned int k=1; k<=colonne; k++)
				cout << it_att[i][k] << "\t";
			cout << endl;
		}
	}
	
	void print_info() {
		cout << "Matrice M(" << righe << "," << colonne << ")" 
		     << " da (" << ux << "," << uy << ") a (" << dx << ","
		     << dy << ")" << endl;
		     
		mostra_risultato();
	}
	
	~BloccoJacobi() {
	}
};

#endif _BLOCCO_JABOBI
