//: blocco_jacobi_par_striscia_din.h
//. Contiene la classe PDBloccoJacobiStriscia, eredita da PBloccoJacobiStriscia.
//. Implementa il protocollo per lo scambio dinamico delle strisce.
#ifndef _BLOCCO_JACOBI_PAR_STRISCIA_DINAMICO
#define _BLOCCO_JACOBI_PAR_STRISCIA_DINAMICO

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include "mpi.h"
#include "blocco_jacobi_par_striscia.h"

using namespace std;	// necessario per utilizzare gli ofstream.

class PDBloccoJacobiStriscia : public PBloccoJacobiStriscia {

	float scarto;	// massimo scarto tollerato tra le velocita'
					// di esecuzione dei processi del gruppo.

	public:
	
	PDBloccoJacobiStriscia(int ux, int uy, int dx, int dy) :
		PBloccoJacobiStriscia(ux, uy, dx, dy) {
		scarto = 0.15;
	}
	
	void set_varianza(unsigned int var) { scarto = var / 100; }
	
	void cambia_matrice(int ux, int uy, int dx, int dy) {
		this->ux = ux;
		this->uy = uy;
		this->dx = dx;
		this->dy = dy;
		
		righe = uy - dy + 1;
		colonne = dx - ux + 1;
		
		proc_start_x = id;
		proc_end_x = id;
		
		if (dx == N) 
			max_x = dx-1;
		else max_x = dx;
		
		if (ux == 1)
			min_x = ux+1;
		else min_x = ux;
		
		start_x = min_x;
		end_x = max_x;

		it_att.resize(righe+1);
		it_new.resize(righe+1);
		if ( (it_att.size()!=righe+1) || (it_new.size()!=righe+1) ) {
			cerr << "Errore durante la creazione del blocco della matricie"
			     << "( (" << ux << "," << uy << "), (" << dx << "," << dy << ") )";
			exit(1);
		}

		for (unsigned int i=0; i<it_att.size(); i++) {
			it_att[i].resize(colonne+1);
			it_new[i].resize(colonne+1);
			if ( (it_att[i].size()!=colonne+1) || (it_new[i].size()!=colonne+1) ) {
				cerr << "Errore durante la creazione del blocco della matricie"
			     << "( (" << ux << "," << uy << "), (" << dx << "," << dy << ") )";
				exit(1);
			}
		}
		
		#ifdef DEBUG
		out << "Matrice cambiata, ora e' " << righe << "x" << colonne << endl;
		out.flush();
		#endif;
	}
	
	unsigned int calcola_jacobi() {
	
		int num_processi;
		
		MPI_Comm_size(MPI_COMM_WORLD, &num_processi);
		
		next_sincronia = 1;
		start_x = min_x;
		end_x = max_x;

		setup_confini();

		MPI_Barrier(MPI_COMM_WORLD);

		tempo_calcolo = MPI_Wtime();
		num_sincronie = 0;
		num_iterazioni = 0;
		
		do {
			// Eseguo una iterazione dell'algoritmo tenendo traccia
			// del tempo impiegato.
			float tempo_iterazione = MPI_Wtime();
			step_jacobi();
			tempo_iterazione = MPI_Wtime() - tempo_iterazione;
			
			// Chiamata collettiva per far ricevere a root i tempi
			// di iterazione di tutti i processi.
			float *tempi = new float[num_processi];
			MPI_Gather(&tempo_iterazione, 1, MPI_FLOAT, tempi, 1, MPI_FLOAT,
				0, MPI_COMM_WORLD);
			#ifdef DEBUG
			out << "Spedito il mio tempo di iterazione: " << tempo_iterazione <<
				endl;
			out.flush();
			#endif
			
			// Valuto se e' il caso di far migrare una striscia ad
			// un altro processo. Lo comunico ai processi spedendo
			// un intero:
			// -1 = devi ridurre il tuo carico di una striscia
			// 1 = devi aumentare il tuo carico di una striscia
			// 10 = devi rieseguire la procedura di sondaggio dei
			// confini.
			// 0 (o qualsiasi altro) = non fare nulla e procedi.
			float *risposte = 0;
			
			if (id == 0) {
				float totale_tempi = 0.0;
				float tempo_massimo = tempi[0];
				unsigned int id_tempo_massimo = 0;
				risposte = new float[num_processi]; // array
						// delle risposte da spedire
						// ai processi.
				
				// Trovo il tempo massimo e ne identifico il
				// processo.
				for (unsigned int i=0; i<num_processi; i++) {
					totale_tempi += tempi[i];
					if (tempi[i] > tempo_massimo) {
						tempo_massimo = tempi[i];
						id_tempo_massimo = i;
					}

					risposte[i] = 0;
				}

				// Calcolo la media dei tempi di esecuzione
				// e lo scarto ammesso.
				float media_tempi = totale_tempi / num_processi;
				float tolleranza = media_tempi * scarto;

				#ifdef DEBUG
				out << "Media tempo iterazione: " << media_tempi << endl;
				out << "Tolleranza: " << tolleranza << endl;
				out.flush();
				#endif

				if (tempo_massimo > media_tempi + tolleranza) {
					for (unsigned int i=0; i<num_processi; i++)
						risposte[i] = 10;

					risposte[id_tempo_massimo] = -1;

					if (id_tempo_massimo < num_processi-1)
						risposte[id_tempo_massimo+1] = 1;
					else
						risposte[id_tempo_massimo-1] = 1;

					#ifdef DEBUG
					out << "Id del processo lento: " << id_tempo_massimo
						<< " con il tempo di: " << tempo_massimo
						<< endl;
					out.flush();
					#endif					
				}
			}
			
			// Spedisco le risposte ai vari processi.
			float risposta;
			MPI_Scatter(risposte, 1, MPI_INT, &risposta, 1, MPI_INT,
				0, MPI_COMM_WORLD);
				
			#ifdef DEBUG
			out << "Ricevuta risposta: " << risposta << endl;
			out.flush();
			#endif
				
			if (risposta == -1 ) {	// devo diminuire il carico di una striscia.
				unsigned int colonna;	// indice della colonna da togliere.
				unsigned int idtemp;	// id del processo a cui la spediro'.
				unsigned int nux = ux;	// nuovo valore di ux.
				unsigned int ndx = dx;	// nuovo valore di dx.
				float *temp = 0;
				
				if (colonne <= 2) {	// se ho meno di 3 colonne non diminuisco il carico.
					temp = new float;
					*temp = 0.0;	// spedisco 0 al processo che sta aspettando la
									// mia colonna per informarlo che non diminuisco
									// il mio carico.
					
					MPI_Send(&temp, 1, MPI_FLOAT, idtemp, 50,
						MPI_COMM_WORLD);
					
					#ifdef DEBUG
					out << "Non tolgo la colonna perche' sono gia' al minimo"
						<< endl;
					out.flush();
					#endif
				} else {
				
					if (id == num_processi-1) {	// se possiedo il confine destro
						colonna = 1;			// levo la striscia del confine sinistro.
						idtemp = id - 1;
						nux++;
					} else {					// ...altrimenti levo la destra.
						colonna = colonne;
						idtemp = id + 1;
						ndx--;
					}
					
					temp = new float[N];	// creo il vettore che contiene la colonna.
					
					for (unsigned int i=1; i<=righe; i++)
						temp[i-1] = it_att[i][colonna];
						
					// spedisco il vettore.
					MPI_Send(temp, N, MPI_FLOAT, idtemp, 50,
						MPI_COMM_WORLD);
						
					#ifdef DEBUG
					out << "Spediti i dati: " << endl;
					for (unsigned int i=0; i<N; i++)
						out << "temp[i]=" << temp[i] << " ";
					out << endl;
					out.flush();
					#endif
					
					// qualora debba togliere la colonna di sinistra devo
					// shiftare le matrici in quanto il metodo cambia_matrice
					// togliera' l'ultima colonna.
					if (colonna == 1)
						for (unsigned int i=1; i<=righe; i++)
							for (unsigned int j=1; j<=colonne-1; j++) {
								it_att[i][j] = it_att[i][j+1];
								it_new[i][j] = it_new[i][j+1];
							}
					
					cambia_matrice(nux, uy, ndx, dy);
					
					#ifdef DEBUG
					out << "Tolta la " << colonna << "a colonna" << endl;
					out.flush();
					print_matrice();
					#endif
				}
				delete []temp;
				
				setup_confini();
			}
			
			if (risposta == 1) {	// devo aumentare il carico di lavoro di una colonna
				float *temp;
				unsigned int colonna;
				unsigned int idtemp;
				unsigned int nux = ux;
				unsigned int ndx = dx;
				int count = 0;
				MPI_Status status;
				
				// ricevo la colonna dal processo straniero.
				temp = new float[N];
				MPI_Recv(temp, N, MPI_FLOAT, MPI_ANY_SOURCE, 50,
					MPI_COMM_WORLD, &status);

				// Se ricevo un solo dato significa che il processo
				// straniero non diminuisce il suo carico.				
				MPI_Get_count(&status, MPI_FLOAT, &count);
				
				if (count != 1) {
				
					#ifdef DEBUG
					out << "Ricevuti " << count << " dati da "
						<< status.MPI_SOURCE << endl;
					for (unsigned int i=0; i<N; i++)
						out << "t[i]=" << temp[i] << " ";
					out << endl;
					#endif
				
					if (status.MPI_SOURCE > id) {
						colonna = colonne + 1;
						ndx++;
					} else {
						colonna = 1;
						nux--;
					}
				
					cambia_matrice(nux, uy, ndx, dy);
				
					if (colonna == 1)
						for (unsigned int i=1; i<=righe; i++)
							for (unsigned int j=colonne; j>=2; j--) {
								it_att[i][j] = it_att[i][j-1];
								it_new[i][j] = it_new[i][j-1];
							}
							
					for (unsigned int i=1; i<=righe; i++) {
						it_att[i][colonna] = temp[i-1];
						it_new[i][colonna] = temp[i-1];
					}
						
					#ifdef DEBUG
					out << "Aggiunta la " << colonna << "a colonna" << endl;
					out.flush();
					print_matrice();
					#endif
				}
				
				delete []temp;
				
				setup_confini();
			}
			
			if (risposta == 10)
				setup_confini();
				
			delete []risposte;
			delete []tempi;
			
			MPI_Barrier(MPI_COMM_WORLD);
			
		} while(errore_globale > convergenza);
		
		tempo_calcolo = MPI_Wtime() - tempo_calcolo;
		return num_iterazioni;
	}
	
	void print_matrice() {
		out << "M ((" << ux << "," << uy << ") (" << dx << "," << dy << "))"
			<< endl;
		for (unsigned int i=1; i<=righe; i++) {
			for (unsigned int k=1; k<=colonne; k++)
				out << it_att[i][k] << "\t";
			out << endl;
		}
	}
};

#endif _BLOCCO_JACOBI_PAR_STRISCIA_DINAMICO
